"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/grouch/lib/__init__.py $
$Id: __init__.py 24750 2004-07-21 15:26:51Z dbinger $

An system for describing and enforcing a Python object schema.
"""

